/*
**                         Copyright 1999 by KVASER AB
**                   P.O Box 4076 SE-51104 KINNAHULT, SWEDEN
**             E-mail: staff@kvaser.se   WWW: http://www.kvaser.se
**
** This software is furnished under a license and may be used and copied
** only in accordance with the terms of such license.
**
** $Header: /SWproj/CKDemo/CKDemoTurin/CKCode.cpp 5     99-11-15 14:53 Erikl $
** $Revision:: 5    $   $NoKeywords: $
**
** Description:
**   Basic implementation of a CAN Kingdom City. We focus on clarity
**   rather than efficiency.
**   The following King's Pages are implemented:
**     0, 1, 2, 3, 4, 16.
**   There are two transmit documents, connected to trackbars in the GUI
**   part, and two receive documents, connected to ProgressBars.
**
**  Environment:
**    Borland C++ Builder 4
**    CANLIB (The virtual channels can be used used so no hardware is needed)
**
**  The program runs under Windows NT and Windows 95/98.
**
*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "CKCode.h"
#include <canlib.h>
#include <shellapi.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

TCity *City;

//---------------------------------------------------------------------------

__fastcall TCity::TCity(TComponent* Owner)
: TForm(Owner)
{
    //
    // In the City constructor the pointers are set up as shown in picture
    // one in the documentation.
    //

    Local_Mayor.Local_Post_Master = &Local_Post_Master;

    Local_Post_Master.Local_Mayor = &Local_Mayor;

    Local_Mayor.Transmit_Document_List[0].Transmit_Receive = TRANSMIT;
    Local_Mayor.Transmit_Document_List[0].Max_Record_No = 3;
    Local_Mayor.Transmit_Document_List[0].Doc[0] = &Mayors_Document;
    Local_Mayor.Transmit_Document_List[0].Doc[1] = &Temp_Sensor_Document_1;
    Local_Mayor.Transmit_Document_List[0].Doc[2] = &Temp_Sensor_Document_2;
    Local_Mayor.Max_Transmit_List_No = 1;

    Local_Mayor.Receive_Document_List[0].Transmit_Receive = RECEIVE;
    Local_Mayor.Receive_Document_List[0].Max_Record_No = 3;
    Local_Mayor.Receive_Document_List[0].Doc[0] = &Kings_Document;
    Local_Mayor.Receive_Document_List[0].Doc[1] = &Temp_Show_Document_1;
    Local_Mayor.Receive_Document_List[0].Doc[2] = &Temp_Show_Document_2;
    Local_Mayor.Max_Receive_List_No = 1;

    Local_Mayor.Max_Folder_No = 6;

    for (int i = 0; i < Local_Mayor.Max_Folder_No; i++) {
        Local_Mayor.Local_Folder[i] = &Local_Folder[i];
    }

    Local_Folder[0].Assigned_Document = &Kings_Document;
    Local_Folder[1].Assigned_Document = &Mayors_Document;

    Local_Folder[0].Enable = true;
    Local_Folder[0].Envelope_Enable = true;
    Local_Folder[0].RTR = false;
    Local_Folder[0].Envelope = 0;
    Local_Folder[0].Transmit_Receive = RECEIVE;

    Local_Folder[1].Enable = true;
    Local_Folder[1].Envelope_Enable = true;
    Local_Folder[1].RTR = false;
    Local_Folder[1].Transmit_Receive = TRANSMIT;
    Kings_Document.Local_Mayor = &Local_Mayor;

    Mayors_Document.Local_Mayor = &Local_Mayor;

    Temp_Sensor_Document_1.Local_Mayor = &Local_Mayor;
    Temp_Sensor_Document_1.Local_Track_Bar = TrackBar1;
    Temp_Sensor_Document_1.Name = "SensorDoc1";

    Temp_Sensor_Document_2.Local_Mayor = &Local_Mayor;
    Temp_Sensor_Document_2.Local_Track_Bar = TrackBar2;
    Temp_Sensor_Document_2.Name = "SensorDoc2";

    Temp_Show_Document_1.Local_Mayor = &Local_Mayor;
    Temp_Show_Document_1.Local_Progress_Bar = ProgressBar1;
    Temp_Show_Document_1.Name = "ShowDoc1";

    Temp_Show_Document_2.Local_Mayor = &Local_Mayor;
    Temp_Show_Document_2.Local_Progress_Bar = ProgressBar2;
    Temp_Show_Document_2.Name = "ShowDoc2";

    //
    // Routines for updating of the City GUI
    //
    InitiateStringGrid1();
    UpDateStringGrid1();
}


/*
** ShowError is called when any Canlib32 API, that should handle the
** communication with the Bus, fails.
*/
void __fastcall TCity::ShowError(char* text, canStatus Stat)
{
    char msg[128], tmp[64];

    canGetErrorText(Stat, tmp, 64);
    strcpy(msg, text);
    strcat(msg, " ");
    strcat(msg, tmp);
    MessageBox(this->Handle, msg, "CAN Library Error", 1);
}

/*
** Sets up the stringgrid.
*/
void __fastcall TCity::InitiateStringGrid1(void)
{
    StringGrid1->Cells [0][1] = "0";
    StringGrid1->Cells [0][2] = "1";
    StringGrid1->Cells [0][3] = "2";
    StringGrid1->Cells [0][4] = "3";
    StringGrid1->Cells [0][5] = "4";
    StringGrid1->Cells [0][6] = "5";

    StringGrid1->Cells [0][0] = "Folder";
    StringGrid1->Cells [1][0] = "Enable";
    StringGrid1->Cells [2][0] = "Doc";
    StringGrid1->Cells [3][0] = "RTR";
    StringGrid1->Cells [4][0] = "Env";
    StringGrid1->Cells [5][0] = "Env Enab.";

    StringGrid1->ColWidths[0] = 45;
    StringGrid1->ColWidths[1] = 50;
    StringGrid1->ColWidths[2] = 65;
    StringGrid1->ColWidths[3] = 30;
    StringGrid1->ColWidths[4] = 55;
    StringGrid1->ColWidths[5] = 70;
}

/*
** Updates the stringgrid(s) and a few labels to show the present
** state of the City.
*/
void __fastcall TCity::UpDateStringGrid1(void)
{
    int i;

    if (Local_Mayor.Base_No > 0) {
        BaseNoLabel->Caption = "BaseNo = " + IntToStr(Local_Mayor.Base_No);
    } else {
        BaseNoLabel->Caption = "BaseNo undefined";
    }

    for (i = 0; i < 6; i++)
    {
        StringGrid1->Cells[1][1+i] = Local_Folder[i].Enable;
        StringGrid1->Cells[3][1+i] = Local_Folder[i].RTR;
        StringGrid1->Cells[5][1+i] = Local_Folder[i].Envelope_Enable;

        if (Local_Folder[i].Envelope == noENVELOPE_ASSIGNED)
            StringGrid1->Cells[4][1+i] = "";
        else
            StringGrid1->Cells[4][1+i] = Local_Folder[i].Envelope;

        if (Local_Folder[i].Assigned_Document != NULL)
            StringGrid1->Cells[2][1+i] =
            Local_Folder[i].Assigned_Document->Name;
        else
            StringGrid1->Cells[2][1+i] = "";
    }

    switch (Local_Post_Master.Communication_Mode)
    {
        case COMM_MODE_KEEP_CURRENT:
            ;//Do nothing
            break;
        case COMM_MODE_SILENT:
            CommLabel->Caption = "Silent";
            break;
        case COMM_MODE_LISTEN_ONLY:
            CommLabel->Caption = "Listen";
            break;
        case COMM_MODE_COMMUNICATE:
            CommLabel->Caption = "Communicate";
            break;
        case COMM_MODE_ERROR:
            CommLabel->Caption = "Error";
            break;
        default: ;
    }
    switch (Local_Mayor.Action_Mode)
    {
        case ACTION_MODE_KEEP_CURRENT:
            ;//Do nothing
            break;
        case ACTION_MODE_RUN:
            ActionLabel->Caption = "Run";
            break;
        case ACTION_MODE_FREEZE:
            ActionLabel->Caption = "Freeze";
            break;
        case ACTION_MODE_RESET:
            ActionLabel->Caption = "Reset";
            break;
        case ACTION_MODE_ERROR:
            ActionLabel->Caption = "Error";
            break;
        default: ;
    }
    for (i = 0; i < 7; i++)
    {
        if (i < Local_Mayor.Local_Groups.Max_No)
            StringGrid2->Cells[i][0] = Local_Mayor.Local_Groups.Group[i];
        else
            StringGrid2->Cells[i][0] = "";
    }
}

/*
** Canlib32 generates an window event when the receivebuffer on the cancontroller
** goes from emty to not emtpy. On the event, the the Local_Post_Master with the
** Receive Message function is called.
*/
void __fastcall TCity::OnCKLetter(TMessage &Message)
{
    Local_Post_Master.ReceiveMessage();
}

/*
** Reacts on a click on the Setup button.
** Calls for the postmaster to setup the CAN circuit.
*/
void __fastcall TCity::setupButtonClick(TObject *Sender)
{
    int i;
    i = ChannelCombo->ItemIndex;
    if (i < 0) {
        // Some CANLIB-related error.
        Local_Post_Master.Channel_Number = -1;
    } else {
        Local_Post_Master.Channel_Number = (int)ChannelCombo->Items->Objects[i];
    }
    Local_Mayor.Physical_Address = (unsigned char)StrToInt(Edit1->Text);
    if (Local_Post_Master.SetupCan(this->Handle, BAUD_125K, 0, 0, 0, 0, 0)) {
        Edit1->Enabled = false;
        ChannelCombo->Enabled = false;
        setupButton->Enabled = false;
    }
}

/*
The TrackBar asks its connected Document for the current Envelope in use, and
passes an empty message with this Envelope to the Local_Mayor's DistributeToFolder
function. Because the Document is a Transmit Document this will result in a
message transmission.
*/
void __fastcall TCity::TrackBar1Change(TObject *Sender)
{
    CKLetter CK_Letter;
    CK_Letter.envelope = Temp_Sensor_Document_1.Envelope;
    CK_Letter.Transmit_Receive = TRANSMIT;
    Local_Mayor.DistributeToFolder(CK_Letter);
}

void __fastcall TCity::TrackBar2Change(TObject *Sender)
{
    CKLetter CK_Letter;
    CK_Letter.envelope = Temp_Sensor_Document_2.Envelope;
    CK_Letter.Transmit_Receive = TRANSMIT;
    Local_Mayor.DistributeToFolder(CK_Letter);
}

//---------------------------------------------------------------------------

/*
** The PostMaster, that handles all the communication with the CAN controller
** (through canlib32 API) is being setup with calls to the proper API's.
*/
int __fastcall PostMaster::SetupCan(
                                     HWND Current_Window,
                                     long Freq,
                                     unsigned int Tseg_1,
                                     unsigned int Tseg_2,
                                     unsigned int Sjw,
                                     unsigned int No_Samp,
                                     unsigned int Sync_Mode)
{
    canStatus stat;

    Handle_To_Opencircuit = canOpenChannel(Channel_Number, canWANT_VIRTUAL);
    if (Handle_To_Opencircuit < 0) {
        City->ShowError("canOpen", (canStatus)Handle_To_Opencircuit);
        return FALSE;
    }

    (void)canSetBusParams(Handle_To_Opencircuit, Freq, Tseg_1,
                          Tseg_2, Sjw, No_Samp, Sync_Mode);

    (void)canSetBusOutputControl(Handle_To_Opencircuit, canDRIVER_NORMAL);

    stat = canBusOn(Handle_To_Opencircuit);
    if (stat != canOK) {
        City->ShowError("canOpen", stat);
        return FALSE;
    }

    // This call will result in a generation of a window event every time
    // the PostMaster's recieve buffer goes from empty to not empty.
    stat = canSetNotify(Handle_To_Opencircuit, Current_Window , canNOTIFY_RX);
    if (stat != canOK) {
        City->ShowError("canSetNotify", stat);
        return FALSE;
    }

    // After setting up the City, the City goes into Freeze Mode waiting and
    // PostMaster into Silent Mode.
    setCommMode(COMM_MODE_SILENT);
    Local_Mayor->setActionMode(ACTION_MODE_FREEZE);

    City->UpDateStringGrid1();
    return TRUE;
}

/*
** CanBus off, circuit closing.
*/
void __fastcall PostMaster::SetDownCan(void)
{
    // Any errors are silently ignored here, as we're closing down anyway.
    (void)canBusOff(Handle_To_Opencircuit);
    (void)canClose(Handle_To_Opencircuit);
    Handle_To_Opencircuit = canINVALID_HANDLE;
}

/*
** The PostMasters function for reading the recive buffer. Because the windows
** event, that triggers this function call, only occurs when the buffer goes from
** empty to not empty, it is necessary to keep reading
** until the queue is emptied.
*/
void __fastcall PostMaster::ReceiveMessage(void)
{
    CKLetter CK_Letter;
    canStatus Stat;

    Stat = canRead(Handle_To_Opencircuit, &CK_Letter.envelope,
                   &CK_Letter.Page, &CK_Letter.Dlc,
                   &CK_Letter.flag, &CK_Letter.time);
    while (Stat == canOK)
    {
        CK_Letter.Transmit_Receive = RECEIVE;
        Local_Mayor->DistributeToFolder(CK_Letter);
        Stat = canRead(Handle_To_Opencircuit,
                       &CK_Letter.envelope, &CK_Letter.Page, &CK_Letter.Dlc,
                       &CK_Letter.flag, &CK_Letter.time);
    }
}

/*
** PostMaster Transmit function.
*/
void __fastcall PostMaster::TransmitMessage(CKLetter CK_Letter)
{
    if (Communication_Mode == COMM_MODE_COMMUNICATE)
    {
        canStatus Stat = canWrite(Handle_To_Opencircuit, CK_Letter.envelope,
                                  CK_Letter.Page, CK_Letter.Dlc,
                                  CK_Letter.flag);
        if (Stat != canOK)
            City->ShowError("canWrite", Stat);
    }
}

/*
** The PostMaster's function for setting the Communication mode.
*/
void __fastcall PostMaster::setCommMode(unsigned char comm_Mode)
{
    switch (comm_Mode)
    {
        case COMM_MODE_KEEP_CURRENT:
            // Do Nothing, keep current mode;
            break;

        case COMM_MODE_SILENT:
            // Silent mode
            // We ignore the return code here as CANLIB might have troubles
            // setting silent mode if there's another program using this
            // channel.
            (void)canSetBusOutputControl(Handle_To_Opencircuit,
                                         canDRIVER_SILENT);
            Communication_Mode = COMM_MODE_SILENT;
            break;

        case COMM_MODE_LISTEN_ONLY:
            //Listen only mode.
            Communication_Mode = COMM_MODE_LISTEN_ONLY;
            break;

        case COMM_MODE_COMMUNICATE:
            // Communicate
            canSetBusOutputControl(Handle_To_Opencircuit,
                                   canDRIVER_NORMAL);
            Communication_Mode = COMM_MODE_COMMUNICATE;
            break;
    }
}

//---------------------------------------------------------------------------

/*
** Searching the Receive and Transmit List for the proper Document.
*/
Document* Mayor::getDocumentByNo(unsigned char Doc_List_No,
                                 unsigned char Doc_Rec_No,
                                 unsigned char Transmit_Receive)
{
    switch (Transmit_Receive)
    {
        case TRANSMIT:
            if (Doc_List_No > Max_Transmit_List_No)
                MessageBox(NULL, "Document list does not exist.", "", 1);
            else
            {
                if (Doc_Rec_No > Transmit_Document_List->Max_Record_No)
                    MessageBox(NULL, "Record does not exist.", "", 1);
                else
                    return Transmit_Document_List[Doc_List_No].Doc[Doc_Rec_No];
            }
            break;
        case RECEIVE:
            if (Doc_List_No > Max_Receive_List_No)
                MessageBox(NULL, "Document list does not exist.", "", 1);
            else
            {
                if (Doc_Rec_No > Receive_Document_List->Max_Record_No)
                    MessageBox(NULL, "Record does not exist.", "", 1);
                else
                    return Receive_Document_List[Doc_List_No].Doc[Doc_Rec_No];
            }
            break;
        default: ;
    }
    return NULL;
}

/*
** The implementation of Reset is, in this environment, mostly a matter of
** taste. Here we expel all the assigned Envelopes, Documents and Folders.
** The Mayor is now in a state identical to where he was just
** after setup, waiting for King's Page 0.
*/
void __fastcall Mayor::Reset(void)
{
    int i;
    Local_Folder[1]->Envelope = noENVELOPE_ASSIGNED;
    Base_No = -1;

    for (i = 2; i < Max_Folder_No; i++)
    {
        Local_Folder[i]->Assigned_Document = NULL;
        Local_Folder[i]->Enable = true;
        Local_Folder[i]->Envelope = noENVELOPE_ASSIGNED;
        Local_Folder[i]->Assigned_Document = NULL;
        Local_Folder[i]->Enable = true;
        Local_Folder[i]->Envelope_Enable = false;
        Local_Folder[i]->RTR = false;
        Local_Folder[i]->Transmit_Receive = RECEIVE;

        Local_Post_Master->setCommMode(COMM_MODE_SILENT);
        setActionMode(ACTION_MODE_FREEZE);
    }
    memset(Local_Groups.Group, 0, 6);
    Local_Groups.Max_No = 0;
}

/*
** Through this function all the Letters are being distributed.
*/
void __fastcall Mayor::DistributeToFolder(CKLetter CK_Letter)
{
    if (CK_Letter.envelope == noENVELOPE_ASSIGNED) return;

    // Handle all Letters in Run mode; in other modes we handle
    // King's Pages only.
    if ((CK_Letter.envelope == 0) || (Action_Mode == ACTION_MODE_RUN))
    {
        for (int i = 0; i < Max_Folder_No; i++)
            Local_Folder[i]->HandleLetter(CK_Letter);
    }
}

/*
** Sets the citys action mode.
*/
void __fastcall Mayor::setActionMode(unsigned char Act_Mode)
{
    Action_Mode = Act_Mode;

    switch (Act_Mode)
    {
        case ACTION_MODE_KEEP_CURRENT:
            //Keep current Mode; Do nothing.
            ;
            break;
        case ACTION_MODE_RUN:
            //run
            break;
        case ACTION_MODE_FREEZE:
            //Freeze
            break;
        case ACTION_MODE_RESET:
            //Reset
            Reset();
            City->UpDateStringGrid1();
            break;
        default: ;
    }
}

//---------------------------------------------------------------------------
/*
** Folder Constructor
*/
__fastcall Folder::Folder()
{
    Assigned_Document = NULL;
    Enable = true;
    Envelope = noENVELOPE_ASSIGNED;
    Assigned_Document = NULL;
    Enable = true;
    Envelope_Enable = false;
    RTR = false;
    Transmit_Receive = RECEIVE;
}

/*
** The Folder's HandleLetter function.  A call to this function with a Letter
** that matches the settings for a Folder will result in a call to the assigned
** Document's handlepage function.
*/
void __fastcall Folder::HandleLetter(CKLetter CK_Letter)
{
/*The condition:(Transmit_Receive == CK_Letter.Transmit_Receive)
   || (RTR == (CK_Letter.flag & 1)) will result in that a recieved message
   with the RTR bit set, will generate a transmission.*/
    if ((Assigned_Document != NULL)
        && (Envelope == CK_Letter.envelope)
        && ((Transmit_Receive == CK_Letter.Transmit_Receive)
            || (RTR == (CK_Letter.flag & 1))
            && (Transmit_Receive == TRANSMIT))
        && (Enable == true)
        && (Envelope_Enable == true))
    {
        Assigned_Document->HandlePage(CK_Letter.Page);
    }
}

//---------------------------------------------------------------------------

/*
** The Base class Document, from which all the Documents are derived.
*/
__fastcall Document::Document()
{
    Envelope = noENVELOPE_ASSIGNED;
    Name = "EmptyDoc";
}

/*Virtual function.*/
void Document::HandlePage(unsigned char* Page){}

//---------------------------------------------------------------------------


/*
** Implementation of the King's Documents.
*/

void KingsDocument::HandlePage(unsigned char* Page)
{
    if ((Local_Mayor->Local_Groups.findGroup(Page[0])
         < Local_Mayor->Local_Groups.Max_No)
        || (Page[0] == Local_Mayor->Physical_Address)
        || (Page[0] == 0))
    {
        switch (Page[1])
        {
            case 0:
                KingsPage0(Page);
                break;
            case 1:
                KingsPage1(Page);
                break;
            case 2:
                KingsPage2(Page);
                break;
            case 3:
                KingsPage3(Page);
                break;
            case 4:
                KingsPage4(Page);
                break;
            case 16:
                KingsPage16(Page);
                break;
            default:
                MessageBox(NULL, "This Kingspage is not implemented.",
                           "Error", 1);
        }
        City->UpDateStringGrid1();
    }
}

/*
** King's Page 0.
** Start/Stop, Modes (Compulsory page)
*/
void __fastcall KingsDocument::KingsPage0(unsigned char* Page)
{
    Local_Mayor->Local_Post_Master->setCommMode(Page[3]);
    Local_Mayor->setActionMode(Page[2]);

    City->UpDateStringGrid1();
}

/*
** King's Page 1.
** Initiate (Base Number, Response request Page) (Compulsory page).
*/
void __fastcall KingsDocument::KingsPage1(unsigned char* Page)
{
    long Base_No = ExtractEnv(Page + 4);
    long Envelope;
    CKLetter CK_Letter;

    if ((Base_No != 0x1fffffff) && (Base_No != 0x0))
    {
        Envelope = Base_No +  Local_Mayor->Physical_Address;
        Local_Mayor->Base_No = Base_No;
        //
        // Sets the Folder one containing MayorsDocument to the specification
        // in this KingsPage.
        //
        Local_Mayor->Local_Folder[1]->Envelope = Envelope;
        Local_Mayor->Local_Folder[1]->Assigned_Document->Envelope = Envelope;
        Local_Mayor->Local_Folder[1]->Envelope_Enable = true;
        Local_Mayor->Local_Folder[1]->Transmit_Receive = TRANSMIT;

        // Passes the Lines in Kingspage1 to the Mayors Document,
        // through a call to DistributeToFolder.
        CK_Letter.envelope = Envelope;
        CK_Letter.Transmit_Receive = TRANSMIT;
    }
    else
        CK_Letter.envelope = Local_Mayor->Local_Folder[1]->Envelope;

    // Send a reply.
    CK_Letter.Page[2] = Page[2];
    Local_Mayor->Local_Folder[1]->HandleLetter(CK_Letter);
}

/*
** King's Page 2. Assign Envelopes (Compulsory).
*/
void __fastcall KingsDocument::KingsPage2(unsigned char* Page)
{
    unsigned char Folder_No = Page[6];
    unsigned char Enable_Disable_Envelope = (uchar)(Page[7] & 1);
    unsigned char Action = (unsigned char)(Page[7] >> 1);
    long Envelope = ExtractEnv(Page + 2);

    Folder *Folder_By_Envelope;
    Folder *Folder_By_FolderNo;

    switch (Action)
    {
        case 0:
            //
            // Keep current Assignment
            //
            Folder_By_Envelope = getFolderByEnvelope(Envelope);

            if (Folder_By_Envelope == NULL)
                MessageBox(NULL, "No such Envelope", "KingPage2", 1);
            else
                Folder_By_Envelope->Envelope_Enable = Enable_Disable_Envelope;
            break;

        case 1:
            //
            // Assign Envelope to Folder. Previous assigned Envelope to this
            // Folder is expelled. (This is an restriction that diverges from the
            // CAN Kingdom specification.)
            //
            Folder_By_FolderNo = getFolderByFolderNo(Folder_No);

            if ((Folder_By_FolderNo == NULL)
                || (Folder_By_FolderNo->Assigned_Document == NULL))
                {
                MessageBox(NULL,
                           "No such Folder and/or no Document in this folder",
                           "KingsPage2", 1);
                }
            else
            {
                Folder_By_FolderNo->Envelope = Envelope;
                Folder_By_FolderNo->Assigned_Document->Envelope = Envelope;
                Folder_By_FolderNo->Envelope_Enable = Enable_Disable_Envelope;
            }

            break;

        case 3:
            //
            // Transfer Envelope to new Folder. The old assignement is canceled
            //
            Folder_By_FolderNo = getFolderByFolderNo(Folder_No);
            Folder_By_Envelope = getFolderByEnvelope(Envelope);

            if ((Folder_By_Envelope == NULL) || (Folder_By_FolderNo == NULL)
                || (Folder_By_FolderNo->Assigned_Document == NULL))
                {
                MessageBox(NULL,
                           "No such Envelope/Folder and/or no "
                           "Document assigned to this Folder",
                           "KingsPage2",  1);
                }
            else
            {
                Folder_By_Envelope->Envelope = noENVELOPE_ASSIGNED;
                Folder_By_Envelope->Assigned_Document->
                                    Envelope = noENVELOPE_ASSIGNED;
                Folder_By_Envelope->Envelope_Enable = false;

                Folder_By_FolderNo->Envelope = Envelope;
                Folder_By_FolderNo->Assigned_Document->Envelope = Envelope;
                Folder_By_Envelope->Envelope_Enable = Enable_Disable_Envelope;
            }
            break;

        case 2:
            //
            // Expel Envelope from any assignment
            //
            Folder_By_Envelope = getFolderByEnvelope(Envelope);

            if ((Folder_By_Envelope != NULL)
                && (Folder_By_Envelope->Assigned_Document != NULL))
                {
                Folder_By_Envelope->Envelope = noENVELOPE_ASSIGNED;
                Folder_By_Envelope->Envelope_Enable = false;
                }
            else
                MessageBox(NULL, "No such Envelope to expel.", "KingsPage2", 1);
            break;

        default:
            MessageBox(NULL, "Unknown error", "KingsPage2", 1);
    }
}

/*
** King's Page 3.
** Assign a City to Groups.
*/
void __fastcall KingsDocument::KingsPage3(unsigned char* Page)
{
    int i = 2;
    while ((Page[i] != 0) && (i < 8))
    {
        Local_Mayor->Local_Groups.addGroup(Page[i]);
        i++;
    }
    City->UpDateStringGrid1();
}

/*
** King's Page 4.
** Remove a City from Groups.
*/
void __fastcall KingsDocument::KingsPage4(unsigned char* Page)
{
    unsigned char i = 2;
    while ((Page[i] != 0) && (i < 8))
    {
        Local_Mayor->Local_Groups.removeGroup(Page[i]);
        i++;
    }
    City->UpDateStringGrid1();
}

/*
** King's Page 16.
** Place Documents into Folders. (Optional Page).
*/
void __fastcall KingsDocument::KingsPage16(unsigned char* Page)
{
    unsigned char Folder_No = Page[2];

    unsigned char RTR = (unsigned char)((Page[3] >> 6) & 1);
    unsigned char Transmit_Receive = (unsigned char)(Page[4] & 1);
    bool Insert_Document = ((Page[4] >> 4) & 1);
    bool Remove_Previous_Document = ((Page[4] >> 5) & 1);
    bool Enable_Folder = ((Page[4] >> 6) & 1);

    unsigned char Document_List_Number = Page[5];
    unsigned char Document_Record_Number = Page[6];

    Folder* Folder_By_FolderNo = getFolderByFolderNo(Folder_No);
    Document* Document_By_No = Local_Mayor->
                               getDocumentByNo(Document_List_Number,
                                                Document_Record_Number,
                                                Transmit_Receive);

    //Checks if the Folder and Document exist.
    if ((Folder_By_FolderNo == NULL) || (Document_By_No == NULL))
        MessageBox(NULL, "The Folder and/or Document does not exist.", "", 1);
    else
    {
        if (Page[3] != 0) //Zeros according to the CAN Kingdom spec means no operation.
        {
            Folder_By_FolderNo->RTR = RTR;
        }
        if (Page[4] != 0) //Zeros according to the CAN Kingdom spec means no operation.
        {
            Folder_By_FolderNo->Enable = Enable_Folder;
            Folder_By_FolderNo->RTR = RTR;
            Folder_By_FolderNo->Transmit_Receive = TRANSMIT;

            if (Insert_Document && Remove_Previous_Document)
            {
                Folder_By_FolderNo->Assigned_Document = Document_By_No;
            }
            else if (!Insert_Document && Remove_Previous_Document)
            {
                Folder_By_FolderNo->Assigned_Document = NULL;
            }
            else if ((Insert_Document) && (!Remove_Previous_Document) &&
                     (Folder_By_FolderNo->Assigned_Document != NULL))
            {
                MessageBox(NULL,
                           "Illegal. A Folder can only contain one document.",
                           "", 1);
            }
            else if ((Insert_Document) && (!Remove_Previous_Document))
            {
                Folder_By_FolderNo->Assigned_Document = Document_By_No;
            }
            City->UpDateStringGrid1();
        }
    }
}

/*
** Search a City Folder by its number.
*/
Folder* KingsDocument::getFolderByFolderNo(unsigned char Folder_No)
{
    if (Folder_No > Local_Mayor->Max_Folder_No)
        return NULL;
    else
        return Local_Mayor->Local_Folder[Folder_No];
}

/*
** Search a City Folder by the Envelopes assigned to it.
*/
Folder* KingsDocument::getFolderByEnvelope(long Envelope)
{
    for (int i = 0; i < Local_Mayor->Max_Folder_No; i++)
    {
        if (Local_Mayor->Local_Folder[i]->Envelope == Envelope)
            return Local_Mayor->Local_Folder[i];
    }
    return NULL;
}

/*
** Orders a part of a message so that it can be read.
*/
long __fastcall KingsDocument::ExtractEnv(unsigned char* msg)
{
    return msg[0] + ((msg[1] + ((msg[2] +((msg[3] & 0x1f) << 8)) << 8)) << 8);
}


//---------------------------------------------------------------------------


/*
** The Mayor's HandlePage function.
*/
void MayorsDocument::HandlePage(unsigned char* Page)
{
    CKLetter CK_Letter;
    CK_Letter.envelope = Envelope;
    CK_Letter.flag = canMSG_STD;
    CK_Letter.Dlc = 8;

    if (Page[2] != 0xff) //0xff :no response is requested.
    {
        switch (Page[2])
        {
            case 0:
                // Send the EAN No
                // In this case, it's hardwired to 733-0130-00000-1
                CK_Letter.Page[0] = 0x00;
                CK_Letter.Page[1] = 0x00; // Page 0
                CK_Letter.Page[2] = 0x40; // EAN, LSB
                CK_Letter.Page[3] = 0xBF;
                CK_Letter.Page[4] = 0xFA;
                CK_Letter.Page[5] = 0xAA;
                CK_Letter.Page[6] = 0xAA; // EAN, MSB
                CK_Letter.Page[7] = 0x00;
                Local_Mayor->Local_Post_Master->TransmitMessage(CK_Letter);
                break;
            case 1:
                // Send the Serial No (here we fake it.)
                CK_Letter.Page[0] = 0x00;
                CK_Letter.Page[1] = 0x01; // Page 1
                CK_Letter.Page[2] = 0xAA; // Serial number, LSB
                CK_Letter.Page[3] = 0xAA;
                CK_Letter.Page[4] = 0xAA;
                CK_Letter.Page[5] = 0xAA;
                CK_Letter.Page[6] = 0xAA; // Serial number, MSB
                CK_Letter.Page[7] = 0;    // Reserved, MBZ.
                Local_Mayor->Local_Post_Master->TransmitMessage(CK_Letter);
                break;
        }
    }
    else
    {
        //Do nothing. No response requested.
    }
}
//---------------------------------------------------------------------------

/*
** Document connected to a trackbar. Checks the bars Position and sends it.
*/
void TempSensorDocument::HandlePage(unsigned char* Page)
{
    CKLetter CK_Letter;
    CK_Letter.envelope = Envelope;
    CK_Letter.Page[0] = (unsigned char)(Local_Track_Bar->Position);
    CK_Letter.flag = canMSG_STD;
    CK_Letter.Dlc = 1;
    Local_Mayor->Local_Post_Master->TransmitMessage(CK_Letter);
}
//---------------------------------------------------------------------------

/*
** Document connected to a ProgressBar. Displays incomming message.
*/
void TempShowDocument::HandlePage(unsigned char* Page)
{
    Local_Progress_Bar->Position = Page[0];
}

//---------------------------------------------------------------------------

void __fastcall Groups::addGroup(unsigned char newGroup)
{
    unsigned char Index = findGroup(newGroup);
    if (!(Index == Max_No)) //Group already there.
        MessageBox(NULL, "The City already belongs to this Group", "", 1);
    else
    {
        Group[Max_No] = newGroup;
        Max_No++;
    }
}

void __fastcall Groups::removeGroup(unsigned char oldGroup)
{
    unsigned char Index = findGroup(oldGroup);
    if (Index == Max_No) //Group not found.
        MessageBox(NULL, "The City doesn't belong to this Group", "", 1);
    else
    {
        Max_No--;
        for (int i = Index; i < Max_No ; i++)
            Group[i] = Group[i + 1];
    }
}

unsigned char __fastcall Groups::findGroup(unsigned char searchGroup)
{
    int i = 0;
    while (i < Max_No && Group[i] != searchGroup)
    {
        i++;
    }
    return (unsigned char)i;
}

void __fastcall TCity::FormShow(TObject *Sender)
{
    int i, n, found;

    // Initialize the CAN routines...
    canInitializeLibrary();

    // ..and enumerate the virtual channels first...
    canGetNumberOfChannels(&n);
    found = 0;
    for (i=0; i<n; i++) {
        int channelType;
        char channelName[64];
        canStatus stat;
        stat = canGetChannelData(i, canCHANNELDATA_CARD_TYPE, &channelType,
                                 sizeof(channelType));
        if (stat != canOK) continue;
        if (channelType != canHWTYPE_VIRTUAL) continue;
        stat = canGetChannelData(i, canCHANNELDATA_CHANNEL_NAME, channelName,
                                 sizeof(channelName));
        ChannelCombo->Items->AddObject((AnsiString)channelName, (TObject*)i);
        found++;
    }
    //
    // ..and then the physical channels; this is a demo and the virtual busses
    // will be used most of the time.
    //
    for (i=0; i<n; i++) {
        int channelType;
        char channelName[64];
        canStatus stat;
        stat = canGetChannelData(i, canCHANNELDATA_CARD_TYPE, &channelType,
                                 sizeof(channelType));
        if (stat != canOK) continue;
        if (channelType == canHWTYPE_VIRTUAL) continue;
        stat = canGetChannelData(i, canCHANNELDATA_CHANNEL_NAME, channelName,
                                 sizeof(channelName));
        ChannelCombo->Items->AddObject((AnsiString)channelName, (TObject*)i);
        found++;
    }

    if (found) ChannelCombo->ItemIndex = 0; else ChannelCombo->ItemIndex = -1;

    // Perform some other GUI setup here.
    CommLabel->Caption = "(Not set up yet)";
    ActionLabel->Caption = "(Not set up yet)";
}
//---------------------------------------------------------------------------

void __fastcall TCity::Button1Click(TObject *Sender)
{
   ShellExecute(NULL, "open", "ckdemoturin.htm", NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

